%  function KernelDimRedTR()
%
%  Computing effective directions for regression with RKHS.
%
%  Author: Kenji Fukumizu (Institute of Statistical Mathematics)
%  Date: March 7, 2006
%
%  (C) Copyright:  Kenji Fukumizu
%
%-----------------------------------------------
% KernelDimRedTR()
%
% Arguments
%  X:  explanatory variables (input data)
%  Y:  response variables (teaching data)
%  K:  dimension of effective subspaces
%  MAX_LOOP: maximum number of iterations
%
% Return value(s)
%  B:  orthonormal column vectors (M x K)
%
% Description
%  This program minimizes the objective function
%     Tr[ Ky(Kz(B)+eps I)^{-1} ] 
%   by the gradient descend method.  
% 
%-----------------------------------------------

function B=KernelDimRedTRG(X,Y,B,MAX_LOOP,SGX,SGY,EPS,ETA,ANL,VERBOSE)

TOL=0.000000001;    % Tolerance for stopping the optimization
INIT_DERIV=0;       % 1: Initialization by derivative method. 0: random 
DISP=1;             % Graphical display for optimization process
RGN=4.8;            % Range for the plot

[N,M]=size(X);  % N: data size, M: dim of X.
L=length(Y(1,:));

% Random initialization of projection matrix 
rand('state',1);
seed=floor(rand(1)*100);
randn('state', seed);

% The projection B is initialized by the derivative method.
if isfloat(B)
    K=B;
    if INIT_DERIV
        fprintf('Initialization by derivative method ... ');
        [B t]=KernelDeriv(X,Y,K,sqrt(ANL)*SGX,SGY,EPS);
        fprintf('done\n');
    else
        B=randn(M,K);
    end
else
    K=length(B(1,:));
end
if K>=M
    fprintf('\nDimension of the effective subspace (K) should be smaller than the dimension of X (M)\n');
    B = 0;
    return;
end
% Normalization
[B,D,V]=svd(B,0);

% Orth. projector to the orthogonal complement to <1>
unit=ones(N,N);
I=eye(N);
Q=I-unit./N;

% Gram matrix of Y
sy2=2*SGY*SGY;
aa=sum(Y.*Y,2);
ab=Y*Y';
D=repmat(aa,1,N);
yy=max(D + D' - 2*ab, zeros(N,N));
Gy=exp(-yy./sy2);  
Kyo=Q*Gy*Q;
Kyo=(Kyo+Kyo')./2;      % Ky(i,j)=Q*exp(-||y(i)-y(j)||^2/SG)*Q

% Initial value of the objective function
Z=X*B;
nZ=Z./sqrt(2)/SGX;
aa=sum(nZ.*nZ,2);
ab=nZ*nZ';
D=repmat(aa,1,N);
ZZ=max(D + D' - 2*ab, zeros(N,N));
Gz=exp(-ZZ);    
Kz=Q*Gz*Q;
Kz=(Kz+Kz')./2;     % Kz(i,j)=Q*exp(-||B^TX(i)-B^TX(j)||^2)*Q

mz=inv(Kz+EPS*N.*I);
tr=sum(sum(Kyo.*mz,1),2);
if VERBOSE
    fprintf('[0]  trace=%.10f\n', tr);    % Initial value 
end

% Main loop for minimization
XXg=zeros(N*M,N);
for a=1:M
    Xa=repmat(X(:,a),1,N);
    XXg((a-1)*N+1:a*N,:)=Xa-Xa';
end

eta=ETA;
pdB=zeros(M,K);   
ssz2=2*SGX*SGX;
ssy2=2*SGY*SGY;

if DISP
    fxy = figure;
    axis([-RGN RGN -RGN RGN]);
    axis square;
    PlotData3class2dim(fxy,X,Y,B);
    drawnow;
end

for h=1:MAX_LOOP
    sz2=ssz2+(ANL-1)*ssz2*(MAX_LOOP-h)/MAX_LOOP;
    sy2=ssy2+(ANL-1)*ssy2*(MAX_LOOP-h)/MAX_LOOP;
    
    Z=X*B;                          
    bb=sum(Z.*Z,2);
    ab=Z*Z';
    Kzw=abs(repmat(bb,1,N) + repmat(bb',N,1) - 2*ab);
    Kzw=exp(-Kzw./sz2);
    mK=mean(Kzw,2);
    rK=repmat(mK,1,N);
    Kz=Kzw-rK-rK'+mean(mK,1)*unit; 
    Kzi=inv(Kz+EPS.*N*eye(N));
    
    Ky=exp(-yy./sy2);
    mK=mean(Ky,2);
    rK=repmat(mK,1,N);
    Ky=Ky-rK-rK'+mean(mK,1)*unit;   
    Kyzi=Ky*Kzi;
    
    ZZg=zeros(N*K,N);
    for b=1:K
        Zb=repmat(Z(:,b),1,N);
        ZZg((b-1)*N+1:b*N,:)=Zb-Zb';
    end
    for a=1:M
        XX=XXg((a-1)*N+1:a*N,:);
        for b=1:K
            tt=XX.*ZZg((b-1)*N+1:b*N,:).*Kzw;
            mK=mean(tt,2);
            rK=repmat(mK,1,N);
            dKB=tt-rK-rK'+mean(mK,1)*unit;        
            dB(a,b)=trace(Kzi*Kyzi*dKB);
        end
    end
    if norm(dB)<TOL 
        break
    end
    % Line search
    nm=norm(dB);   
    [B tr] = KDRline(X,Ky,sz2,B,dB/nm,eta,EPS);
    [B,D,V]=svd(B,0);
    
    % Computatoin of the trace with unanealed parameter
    if VERBOSE
        sy2=2*SGY*SGY;
        aa=sum(Y.*Y,2);
        ab=Y*Y';
        D=repmat(aa,1,N);
        yy=max(D + D' - 2*ab, zeros(N,N));
        Gy=exp(-yy./sy2);  
        Kyo=Q*Gy*Q;
        Kyo=(Kyo+Kyo')./2;  
        Z=X*B;
        nZ=Z./sqrt(2)/SGX;
        aa=sum(nZ.*nZ,2);
        ab=nZ*nZ';
        D=repmat(aa,1,N);
        ZZ=max(D + D' - 2*ab, zeros(N,N));
        Gz=exp(-ZZ);    
        Kz=Q*Gz*Q;
        Kz=(Kz+Kz')./2;     % Kz(i,j)=Q*exp(-||B^TX(i)-B^TX(j)||^2)*Q

        mz=inv(Kz+EPS*N.*I);
        tr=sum(sum(Kyo.*mz,1),2);

        fprintf('[%d]  tr=%.10f\n', h, tr);
        
        
    end
    if DISP     
        PlotData3class2dim(fxy,X,Y,B);
        drawnow;
    end
end




